import inspect
from typing import Dict, Any, TypedDict, get_type_hints
from langgraph.graph import StateGraph, START, END
import os


class Davia:
    """
    Main application class that hold all registered subobjects
    """

    def __init__(self, name: str = "davia"):
        self.name = name
        self.tasks = {}
        self.graphs = {}

    @property
    def task(self):
        """
        Decorator to register a task.
        Usage:
            @app.task
            def my_task():
                pass
        """

        def decorator(func):
            # Get source file information
            source_file = inspect.getsourcefile(func)
            if source_file:
                source_file = os.path.relpath(source_file)

            # Store graph with metadata
            self.tasks[func.__name__] = {
                "source_file": source_file,  # Store the source file
            }

            # Return the original function
            return func

        return decorator

    @property
    def graph(self):
        """
        Decorator to register a graph.
        Usage:
            @app.graph
            def my_graph():
                graph = StateGraph(State)
                graph.add_node("node", node_func)
                graph.add_edge(START, "node")
                graph.add_edge("node", END)
                return graph
        """

        def decorator(func):
            # Get source file information
            source_file = inspect.getsourcefile(func)
            if source_file:
                source_file = os.path.relpath(source_file)

            # Store graph with metadata
            self.graphs[func.__name__] = {
                "source_file": source_file,  # Store the source file
            }

            # Return the graph instance for direct access
            return func

        return decorator

    def list_tasks(self):
        """
        List all registered tasks
        """
        return list(self.tasks.keys())

    def list_graphs(self):
        """
        List all registered graphs
        """
        return list(self.graphs.keys())

    def get_task_info(self, task_name: str) -> Dict[str, Any]:
        """
        Get detailed information about a task including its parameters and docstring.
        """
        if task_name not in self.tasks:
            raise KeyError(f"Task '{task_name}' not found")
        return self.tasks[task_name]

    def get_graph_info(self, graph_name: str) -> Dict[str, Any]:
        """
        Get detailed information about a graph including its parameters and docstring.
        """
        if graph_name not in self.graphs:
            raise KeyError(f"Graph '{graph_name}' not found")
        return self.graphs[graph_name]



import typer
from rich import print
from typing_extensions import Annotated
from pathlib import Path
import importlib.util
import sys
from davia.langgraph.launcher import run_server
from davia.app.application import Davia

app = typer.Typer(no_args_is_help=True, rich_markup_mode="markdown")


@app.callback()
def callback():
    """
    :sparkles: Davia
    - Customize your UI with generative components
    - Experience the perfect fusion of human creativity and artificial intelligence!
    - Get started here: [quickstart](https://docs.davia.ai/quickstart)
    """


def _load_app_from_file(file_path: Path) -> Davia:
    """Load a Davia app from a Python file."""
    # Convert to absolute path
    file_path = file_path.absolute()

    # Create a module spec
    spec = importlib.util.spec_from_file_location("davia_app", file_path)
    if spec is None or spec.loader is None:
        raise ImportError(f"Could not load module from {file_path}")

    # Add the parent directory to sys.path to allow relative imports
    sys.path.insert(0, str(file_path.parent))

    # Load the module
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Find the Davia app instance
    for name, obj in module.__dict__.items():
        if isinstance(obj, Davia):
            return obj

    raise ValueError(f"No Davia app instance found in {file_path}")


@app.command()
def run(
    file: Annotated[
        Path,
        typer.Argument(
            help="Path to a Python file containing a Davia app. The file should contain a Davia app instance."
        ),
    ],
    host: Annotated[
        str,
        typer.Option(
            help="The host to serve on. For local development use [blue]127.0.0.1[/blue]. For public access use [blue]0.0.0.0[/blue]."
        ),
    ] = "127.0.0.1",
    port: Annotated[
        int,
        typer.Option(help="The port to serve on."),
    ] = 2025,
    reload: Annotated[
        bool,
        typer.Option(
            help="Enable auto-reload of the server when files change. Use only during development."
        ),
    ] = True,
    n_jobs_per_worker: Annotated[
        int,
        typer.Option(help="Number of jobs per worker."),
    ] = 1,
    browser: Annotated[
        bool,
        typer.Option(help="Open browser automatically when server starts."),
    ] = True,
):
    """
    Run a Davia app from a Python file.

    The file should contain a Davia app instance that will be used to run the server.
    """
    try:
        app = _load_app_from_file(file)
        run_server(
            app=app,
            host=host,
            port=port,
            reload=reload,
            n_jobs_per_worker=n_jobs_per_worker,
            browser=browser,
        )
    except Exception as e:
        print(f"[red]Error: {str(e)}[/red]")
        raise typer.Exit(1)



import uvicorn
from pathlib import Path
import json
from rich import print
from dotenv import load_dotenv
from langgraph_api.cli import patch_environment
import typer
import threading

from davia.app.application import Davia


def run_server(
    app: Davia,
    host: str = "127.0.0.1",
    port: int = 2025,
    n_jobs_per_worker: int = 1,
    browser: bool = True,
    reload: bool = True,
):
    local_url = f"http://{host}:{port}"
    preview_url = "https://dev.davia.ai/dashboard"

    def _open_browser():
        import time
        import urllib.request

        while True:
            try:
                with urllib.request.urlopen(f"{local_url}/info") as response:
                    if response.status == 200:
                        typer.launch(preview_url)
                        return
            except urllib.error.URLError:
                pass
            time.sleep(0.1)

    if browser:
        threading.Thread(target=_open_browser, daemon=True).start()

    print(f"""
        Welcome to
▗▄▄▄   ▗▄▖ ▗▖  ▗▖▗▄▄▄▖ ▗▄▖ 
▐▌  █ ▐▌ ▐▌▐▌  ▐▌  █  ▐▌ ▐▌
▐▌  █ ▐▛▀▜▌▐▌  ▐▌  █  ▐▛▀▜▌
▐▙▄▄▀ ▐▌ ▐▌ ▝▚▞▘ ▗▄█▄▖▐▌ ▐▌

- 🎨 UI: {preview_url}
""")

    filtered_graphs = {
        name: f"{Path(graph_data['source_file']).resolve().as_posix()}:{name}"
        for name, graph_data in app.graphs.items()
    }
    tasks = app.tasks

    # Get the absolute path to launcher_graph.py
    current_file_path = Path(__file__).resolve()
    # Get the directory containing launcher_graph.py
    current_dir = current_file_path.parent
    # Get the absolute path to custom_app.py in the same directory
    custom_app_path = current_dir / "custom_app.py"

    http = {"app": f"{custom_app_path.resolve().as_posix()}:app"}

    with patch_environment(
        MIGRATIONS_PATH="__inmem",
        DATABASE_URI=":memory:",
        REDIS_URI="fake",
        N_JOBS_PER_WORKER=str(n_jobs_per_worker if n_jobs_per_worker else 1),
        LANGSERVE_GRAPHS=json.dumps(filtered_graphs) if filtered_graphs else None,
        TASKS=json.dumps(tasks) if tasks else None,
        LANGSMITH_LANGGRAPH_API_VARIANT="local_dev",
        LANGGRAPH_HTTP=json.dumps(http) if http else None,
        # See https://developer.chrome.com/blog/private-network-access-update-2024-03
        ALLOW_PRIVATE_NETWORK="true",
    ):
        load_dotenv()
        uvicorn.run(
            "langgraph_api.server:app",
            host=host,
            port=port,
            reload=reload,
            log_level="warning",
            access_log=False,
            log_config={
                "version": 1,
                "incremental": False,
                "disable_existing_loggers": False,
                "formatters": {
                    "simple": {
                        "class": "langgraph_api.logging.Formatter",
                    }
                },
                "handlers": {
                    "console": {
                        "class": "logging.StreamHandler",
                        "formatter": "simple",
                        "stream": "ext://sys.stdout",
                    }
                },
                "root": {"handlers": ["console"]},
            },
        )
